﻿
'A simple VB.net (2008) application to demonstrate the use of
'MIDI in ports, MIDI out ports and the MIDI stream buffer.

'If you find any bugs or make any significant improvements
'please send me a copy of your revised code

'Author : J.O.Linton
'Contact: jolinton@btinternet.com

'Version: 1.00
'Date: April 2008

Imports System.Runtime.InteropServices

Public Class Form1

  Private MidiOutPort As New MIDIOutputPortType

  'must include withevents keyword here so that callback function can raise an event
  Private WithEvents MidiInPort As New MidiInputPortType
  Private WithEvents MidiStreamPort As New MIDIOutputStreamType

  ' This delegate enables asynchronous calls for setting
  ' the text property on a TextBox control.
  Delegate Sub SetTextCallback(ByVal text As String)

  'dimension two MIDIHDRs to implement leapfrogging
  Private Mheader(1) As MIDIHDR
  Private BufferSize As Integer = 1200 ' holds up to 100 midi events

  'event list (for demo purposes only)
  Private EventList(,) As Integer
  Private Nevents As Integer
  Private index As Integer

  Private Sub Form1_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
    MidiOutPort.Close()
    MidiInPort.Close()
    MidiStreamPort.Close()
  End Sub

  Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load

    'give the two MIDIHDRs a unique ID
    Mheader(0).MhdrID = 0
    Mheader(1).MhdrID = 1

    'allocate memory space for the midi event buffers
    Mheader(0).DataPtr = Marshal.AllocHGlobal(BufferSize)
    Mheader(1).DataPtr = Marshal.AllocHGlobal(BufferSize)
    Mheader(0).BufferLength = BufferSize
    Mheader(1).BufferLength = BufferSize

    Dim NumberOfInputDevices As Integer = API.NumInDevices
    If NumberOfInputDevices = 0 Then
      MsgBox("No MIDI Input ports available")
    Else
      ComboBox1.Items.Clear()
      For device = 0 To NumberOfInputDevices - 1
        MidiInPort.Open(device)
        ComboBox1.Items.Add(MidiInPort.Info.Label)
        MidiInPort.Close()
      Next
    End If

    Dim NumberOfOutputDevices As Integer = API.NumOutDevices
    If NumberOfOutputDevices = 0 Then
      MsgBox("No MIDI Output ports available")
    Else
      ComboBox2.Items.Clear()
      ComboBox3.Items.Clear()
      For device = 0 To NumberOfOutputDevices - 1
        MidiOutPort.Open(device)
        ComboBox2.Items.Add(MidiOutPort.Info.Label)
        ComboBox3.Items.Add(MidiOutPort.Info.Label)
        MidiOutPort.Close()
      Next
    End If

    'set up the events needed for a simple tune consisting of three crotchet chords
    Nevents = 14
    ReDim EventList(Nevents, 1)
    EventList(0, 0) = 0 : EventList(0, 1) = (MEVT_TEMPO << 24) Or 200000
    EventList(1, 0) = 0 : EventList(1, 1) = (MEVT_NOP) << 24 ' this is just to try it
    EventList(2, 0) = 0 : EventList(2, 1) = &H604790
    EventList(3, 0) = 0 : EventList(3, 1) = &H604A90
    EventList(4, 0) = 96 : EventList(4, 1) = &H604780
    EventList(5, 0) = 0 : EventList(5, 1) = &H604A80
    EventList(6, 0) = 0 : EventList(6, 1) = &H604590
    EventList(7, 0) = 0 : EventList(7, 1) = &H604890
    EventList(8, 0) = 96 : EventList(8, 1) = &H604580
    EventList(9, 0) = 0 : EventList(9, 1) = &H604880
    EventList(10, 0) = 0 : EventList(10, 1) = &H604390
    EventList(11, 0) = 0 : EventList(11, 1) = &H604790
    EventList(12, 0) = 96 : EventList(12, 1) = &H604380
    EventList(13, 0) = 0 : EventList(13, 1) = &H604780
  End Sub

  Private Sub ComboBox1_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ComboBox1.SelectedIndexChanged
    MidiInPort.Close()
    Dim err As Integer = MidiInPort.Open(ComboBox1.SelectedIndex)
    If err Then
      MsgBox(MidiInPort.GetErrorText(err))
      ComboBox1.BackColor = Color.LightGray
    Else
      ComboBox1.BackColor = Color.White
      MidiInPort.StartRecording()
    End If
  End Sub

  Private Sub ComboBox2_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ComboBox2.SelectedIndexChanged
    MidiOutPort.Close()
    Dim err As Integer = MidiOutPort.Open(ComboBox2.SelectedIndex)
    If err Then
      MsgBox(MidiOutPort.GetErrorText(err))
      ComboBox2.BackColor = Color.LightGray
    Else
      ComboBox2.BackColor = Color.White
    End If
  End Sub

  Private Sub ComboBox3_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ComboBox3.SelectedIndexChanged
    MidiStreamPort.Close()
    Dim err As Integer = MidiStreamPort.Open(ComboBox3.SelectedIndex)
    If err Then MsgBox(MidiStreamPort.GetErrorText(err))
    If err Then
      MsgBox(MidiStreamPort.GetErrorText(err))
      ComboBox3.BackColor = Color.LightGray
    Else
      ComboBox3.BackColor = Color.White
    End If
  End Sub

  Private Sub Button123_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Button1.Click, Button2.Click, Button3.Click
    If MidiOutPort.DeviceHandle = 0 Then MsgBox("Please select an output device.") : Exit Sub
    Dim pitch As Integer
    Select Case True
      Case sender Is Button1 : pitch = &H42
      Case sender Is Button2 : pitch = &H40
      Case sender Is Button3 : pitch = &H3E
    End Select
    MidiOutPort.SendMsg(&H600090 Or (pitch << 8)) 'note on
    Wait(500)
    MidiOutPort.SendMsg(&H600080 Or (pitch << 8)) 'note off
  End Sub

  Private Sub Wait(ByVal TimeToWait)
    Dim t As Integer = My.Computer.Clock.TickCount + TimeToWait
    Do While My.Computer.Clock.TickCount < t
    Loop
  End Sub

  Private Sub Button4_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Button4.Click
    If MidiStreamPort.DeviceHandle = 0 Then MsgBox("Please select an output device.") : Exit Sub
    MidiStreamPort.StopStream() ' reset the clock to zero
    index = 0
    PrepareMidiData(0) 'prepare the buffer for the first crotchet on Mheader 0
    PrepareMidiData(1) 'prepare the buffer for the second crotchet on Mheader 1
    MidiStreamPort.RestartStream() ' start the clock
    OutputMidiData(0) ' sound the first crotchet
  End Sub

  Private Sub PrepareMidiData(ByVal ID As Integer)
    If index > Nevents Then Exit Sub

    'this is the way it is supposed to be done but it seems unnecessary
    'in any case it is not easy to add more than on event into the buffer
    'as you are not allowed to add an integer to the IntPtr in the StructureToPtr call

    'Dim Mevent As New THREEBYTE_MIDIEVENT
    'Dim Size As Integer = Marshal.SizeOf(Mevent)
    'Mevent.DeltaTime = EventList(index, 0)
    'Mevent.StreamID = 0
    'Mevent.Message = EventList(index, 1)
    'Marshal.StructureToPtr(Mevent, Mheader(MhdrID).DataPtr, False)
    'Mheader(MhdrID).BytesRecorded = Size

    'instead I prefer to write the integers directly into the buffer
    Dim bytes As Integer = 0
    Do
      Marshal.WriteInt32(Mheader(ID).DataPtr, bytes, EventList(index, 0))
      Marshal.WriteInt32(Mheader(ID).DataPtr, bytes + 4, 0)
      Marshal.WriteInt32(Mheader(ID).DataPtr, bytes + 8, EventList(index, 1))
      bytes += 12 : index += 1
    Loop Until index > Nevents OrElse EventList(index, 0) > 0 Or bytes = BufferSize
    Mheader(ID).BytesRecorded = bytes
    Mheader(ID).BufferLength = BufferSize

    Dim err As Integer = MidiStreamPort.PrepareHeader(Mheader(ID))
    If err Then MsgBox(MidiStreamPort.GetErrorText(err))
  End Sub

  Private Sub OutputMidiData(ByVal ID)
    Dim err As Integer = MidiStreamPort.OutputEvents(Mheader(ID))
    If err Then MsgBox(MidiStreamPort.GetErrorText(err))
  End Sub

  Private Sub UnprepareMidiData(ByVal ID)
    Dim err As Integer = MidiStreamPort.UnPrepareHeader(Mheader(ID))
    If err Then MsgBox(MidiStreamPort.GetErrorText(err))
  End Sub

  Private Sub MidiHeaderDone(ByVal ID) Handles MidiStreamPort.MidiHdrDone
    'NB we must prepare the next header BEFORE outputting the one that has already been prepared
    'because some headers may be played immediately and the callback arrives too soon
    UnprepareMidiData(ID)
    PrepareMidiData(ID)
    Dim otherheader As Integer = (ID + 1) Mod 2
    If Mheader(otherheader).Flags And MHDR_PREPARED Then
      OutputMidiData(otherheader) 'output the other header
    End If
  End Sub

  Private Sub NoteRecorded(ByVal msg As Integer, ByVal timestamp As Integer) Handles MidiInPort.MidiMsgArrived
    'NB textbox1.text must be set using a thread-safe method
    If (msg And &HF0) = &H90 Then SetText(Str(msg >> 8 And &HFF))
    If (msg And &HF0) = &H90 Or (msg And &HF0 = &HB0) Then
      'echo note on and note off messages
      If CheckBox1.Checked Then MidiOutPort.SendMsg(msg)
    End If
  End Sub

  Private Sub SetText(ByVal text As String)
    ' InvokeRequired required compares the thread ID of the
    ' calling thread to the thread ID of the creating thread.
    ' If these threads are different, it returns true.
    If Me.TextBox1.InvokeRequired Then
      Dim d As New SetTextCallback(AddressOf SetText)
      Me.Invoke(d, New Object() {text})
    Else
      Me.TextBox1.Text = text
    End If
  End Sub
End Class
