﻿
'MIDI module implementing MIDI input port, MIDI output port and MIDI stream buffer

'Author : J.O.Linton
'Contact: jolinton@btinternet.com

Imports System.Runtime.InteropServices

Module MIDIModule
  Public Const CALLBACK_FUNCTION = &H30000
  Public Const CALLBACK_NULL = &H0
  Public Const MEVT_SHORTMSG = &H0
  Public Const MEVT_TEMPO = &H1
  Public Const MEVT_NOP = &H2
  Public Const MOM_OPEN = &H3C7
  Public Const MOM_CLOSE = &H3C8
  Public Const MOM_DONE = &H3C9
  Public Const MHDR_DONE = &H1
  Public Const MHDR_PREPARED = &H2
  Public Const MHDR_INQUEUE = &H4
  Public Const MHDR_ISSTRM = &H8

  <StructLayout(LayoutKind.Sequential)> Public Structure MIDIHDR
    Dim DataPtr As IntPtr ' pointer to MIDI data
    Dim BufferLength As Integer ' size of buffer
    Dim BytesRecorded As Integer ' actual amount of data in buffer
    Dim MhdrID As Integer 'custom user data
    Dim Flags As Integer 'flags giving information about the buffer (see below)
    Dim Reserved1 As Integer
    Dim Reserved2 As Integer
    Dim CallbackOffset As Integer ' offset into buffer when callback is performed
    Dim Reserved3a As Integer ' contrary to what is implied in the documentation,
    Dim Reserved3b As Integer ' 4 integers appear to be required here
    Dim Reserved3c As Integer ' ditto
    Dim Reserved3d As Integer ' ditto
  End Structure

  <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> Public Structure MIDIINCAPS
    Dim ManufacturerID As Short
    Dim ProductID As Short
    Dim DriverVersion As Integer
    <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=32)> Dim Label As String
    Dim Support As Integer
  End Structure

  <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> Public Structure MIDIOUTCAPS
    Dim ManufacturerID As Short
    Dim ProductID As Short
    Dim DriverVersion As Integer
    <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=32)> Dim Label As String
    Dim Technology As Short
    Dim Voices As Short
    Dim Notes As Short
    Dim ChannelMask As Short
    Dim Support As Integer
  End Structure

  <StructLayout(LayoutKind.Sequential)> Structure THREEBYTE_MIDIEVENT
    Dim DeltaTime As Integer '  Ticks since last event
    Dim StreamID As Integer '  Reserved; must be zero
    Dim Message As Integer '  MIDI data. MSB must be zero
  End Structure

  Public Class API

    'To see how API functions are called from VB.net
    'http://msdn2.microsoft.com/en-us/library/172wfck9(VS.80).aspx
    'is essential reading

    'not all these calls require the use of the auto modifier but all work with it

    'MIDI Input functions
    Declare Auto Function midiInGetNumDevs Lib "winmm.dll" () As Integer
    Declare Auto Function midiInGetDevCaps Lib "winmm.dll" (ByVal DeviceID As Integer, ByRef DevCaps As MIDIINCAPS, ByVal Size As Integer) As Integer
    Declare Auto Function midiInOpen Lib "winmm.dll" (ByRef DeviceHandle As Integer, ByVal DeviceID As Integer, ByVal CallbackPtr As [Delegate], ByVal Instance As Integer, ByVal CallbackFlags As Integer) As Boolean
    Declare Auto Function midiInClose Lib "winmm.dll" (ByVal DeviceHandle As Integer) As Integer
    Declare Auto Function midiInStart Lib "winmm.dll" (ByVal DeviceHandle As Integer) As Integer
    Declare Auto Function midiInStop Lib "winmm.dll" (ByVal DeviceHandle As Integer) As Integer

    'MIDI Output functions
    Declare Auto Function midiOutGetNumDevs Lib "winmm.dll" () As Integer
    'NB DevCaps must be called ByRef
    Declare Auto Function midiOutGetDevCaps Lib "winmm.dll" (ByVal DeviceID As Integer, ByRef DevCaps As MIDIOUTCAPS, ByVal Size As Integer) As Integer
    'NB DeviceHandle must be called ByRef
    Declare Auto Function midiOutOpen Lib "winmm.dll" (ByRef DeviceHandle As Integer, ByVal DeviceID As Integer, ByVal cbfunc As Integer, ByVal cbdata As Integer, ByVal cboptions As Integer) As Integer
    Declare Auto Function midiOutClose Lib "winmm.dll" (ByVal DeviceHandle As Integer) As Integer
    Declare Auto Function midiOutShortMsg Lib "winmm.dll" (ByVal DeviceHandle As Integer, ByVal msg As Integer) As Integer

    'MIDI Stream buffer functions
    'NB BOTH StreamHandle AND DeviceID must be called ByRef
    Declare Auto Function midiStreamOpen Lib "winmm.dll" (ByRef StreamHandle As Integer, ByRef DeviceID As Integer, ByVal cMidi As Integer, ByVal CallbackPtr As [Delegate], ByVal Instance As Integer, ByVal CallbackFlags As Integer) As Integer
    'NB cMidi must = 1
    Declare Auto Function midiStreamClose Lib "winmm.dll" (ByVal StreamHandle As Integer) As Integer
    'NB Hdr must be called ByRef so that flags can be updated
    Declare Auto Function midiStreamOut Lib "winmm.dll" (ByVal StreamHandle As Integer, ByRef Hdr As MIDIHDR, ByVal size As Integer) As Integer
    Declare Auto Function midiStreamPause Lib "winmm.dll" (ByVal StreamHandle As Integer) As Integer
    Declare Auto Function midiStreamRestart Lib "winmm.dll" (ByVal StreamHandle As Integer) As Integer
    Declare Auto Function midiStreamStop Lib "winmm.dll" (ByVal StreamHandle As Integer) As Integer

    'NB Hdr must be called ByRef  so that flags can be updated
    Declare Auto Function midiOutPrepareHeader Lib "winmm.dll" (ByVal Handle As Integer, ByRef Hdr As MIDIHDR, ByVal Size As Integer) As Integer
    'NB Hdr must be called ByRef  so that flags can be updated
    Declare Auto Function midiOutUnprepareHeader Lib "winmm.dll" (ByVal Handle As Integer, ByRef Hdr As MIDIHDR, ByVal Size As Integer) As Integer

    'other functions
    Declare Auto Function midiInGetErrorText Lib "winmm.dll" (ByVal ErrorCode As Integer, ByVal Text As String, ByVal TextLength As Integer) As Integer
    Declare Auto Function midiOutGetErrorText Lib "winmm.dll" (ByVal ErrorCode As Integer, ByVal Text As String, ByVal TextLength As Integer) As Integer


    Public Shared Function NumInDevices() As Integer
      'returns the no. of MIDI input devices installed in the system
      Return API.midiInGetNumDevs
    End Function
    Public Shared Function NumOutDevices() As Integer
      'returns the no. of MIDI output devices installed in the system
      Return API.midiOutGetNumDevs
    End Function

  End Class

  Public Class MidiInputPortType

    Private Delegate Sub CallbackPtr(ByVal a As Integer, ByVal b As Integer, ByVal c As Integer, ByVal d As Integer, ByVal e As Integer)

    Public Info As MIDIINCAPS
    Public DeviceHandle As Integer
    Private CallbackProcedurePtr As CallbackPtr = AddressOf CallbackProcedure
    Event MidiMsgArrived(ByVal msg As Integer, ByVal timestamp As Integer)

    Public Function Open(ByVal DeviceID)
      API.midiInGetDevCaps(DeviceID, Info, Marshal.SizeOf(Info))
      Return API.midiInOpen(DeviceHandle, DeviceID, CallbackProcedurePtr, 0, CALLBACK_FUNCTION)
    End Function

    Public Sub Close()
      If DeviceHandle <> 0 Then API.midiInClose(DeviceHandle)
      DeviceHandle = 0
    End Sub

    Public Sub StartRecording()
      API.midiInStart(DeviceHandle)
    End Sub

    Public Sub StopRecording()
      If DeviceHandle <> 0 Then API.midiInStop(DeviceHandle)
    End Sub

    Private Sub CallbackProcedure(ByVal DeviceHandle, ByVal MidiMsg, ByVal Instance, ByVal Param1, ByVal Param2)
      RaiseEvent MidiMsgArrived(Param1, Param2)
    End Sub

    Public Function GetErrorText(ByVal ErrorCode)
      Dim text As String = Space(128)
      API.midiInGetErrorText(ErrorCode, text, 128)
      Return text
    End Function

  End Class

  Public Class MIDIOutputPortType

    Public Info As MIDIOUTCAPS
    Public DeviceHandle As Integer 'Handle to the midi device

    Public Function Open(ByVal DeviceId)
      API.midiOutGetDevCaps(DeviceId, Info, Marshal.SizeOf(Info))
      Return API.midiOutOpen(DeviceHandle, DeviceId, 0, 0, CALLBACK_NULL)
    End Function

    Public Sub Close()
      If DeviceHandle <> 0 Then API.midiOutClose(DeviceHandle)
      DeviceHandle = 0
    End Sub

    Public Sub SendMsg(ByVal msg)
      If DeviceHandle <> 0 Then API.midiOutShortMsg(DeviceHandle, msg)
    End Sub

    Public Function GetErrorText(ByVal ErrorCode)
      Dim text As String = Space(128)
      API.midiOutGetErrorText(ErrorCode, text, 128)
      Return text
    End Function

  End Class

  Public Class MIDIOutputStreamType

    Private Delegate Sub CallbackPtr(ByVal a As Integer, ByVal b As Integer, ByVal c As Integer, ByRef d As MIDIHDR, ByVal e As Integer)

    Public Info As MIDIOUTCAPS
    Public DeviceHandle As Integer = 0 'Handle to the midi device 
    Private CallbackProcedurePtr As CallbackPtr = AddressOf CallbackProcedure
    Event MidiHdrDone(ByVal MhdrID As Integer)

    Public Function Open(ByVal DeviceId)
      API.midiOutGetDevCaps(DeviceId, Info, Marshal.SizeOf(Info))
      Return API.midiStreamOpen(DeviceHandle, DeviceId, 1, CallbackProcedurePtr, 0, CALLBACK_FUNCTION)
    End Function

    Public Sub Close()
      If DeviceHandle <> 0 Then API.midiStreamClose(DeviceHandle)
      DeviceHandle = 0
    End Sub

    Public Function PrepareHeader(ByRef data As MIDIHDR)
      Return API.midiOutPrepareHeader(DeviceHandle, data, Marshal.SizeOf(data))
    End Function

    Public Function OutputEvents(ByRef data As MIDIHDR)
      Return API.midiStreamOut(DeviceHandle, data, Marshal.SizeOf(data))
    End Function

    Public Function UnPrepareHeader(ByRef data As MIDIHDR)
      Return API.midiOutUnprepareHeader(DeviceHandle, data, Marshal.SizeOf(data))
    End Function

    Public Sub RestartStream()
      API.midiStreamRestart(DeviceHandle)
    End Sub

    Public Sub StopStream()
      If DeviceHandle <> 0 Then API.midiStreamStop(DeviceHandle)
    End Sub

    Public Sub SendMsg(ByVal msg)
      If DeviceHandle <> 0 Then API.midiOutShortMsg(DeviceHandle, msg)
    End Sub

    Private Sub CallbackProcedure(ByVal DeviceHandle As Integer, ByVal MidiMsg As Integer, ByVal Instance As Integer, ByRef MHdr As MIDIHDR, ByVal Param2 As Integer)
      Select Case MidiMsg
        Case MOM_OPEN
        Case MOM_DONE : RaiseEvent MidiHdrDone(MHdr.MhdrID)
        Case MOM_CLOSE
      End Select
    End Sub

    Public Function GetErrorText(ByVal ErrorCode)
      Dim text As String = Space(128)
      API.midiOutGetErrorText(ErrorCode, text, 128)
      Return text
    End Function

  End Class

End Module
